// src/frontend/pause.ts
function is_paused(pause_reasons) {
  return Object.values(pause_reasons).some((pause) => pause);
}
function pause(pause_reasons, motive = "pause") {
  pause_reasons[motive] = true;
}
function resume(pause_reasons, motive) {
  if (!motive || motive.length == 0) {
    return Object.keys(pause_reasons).forEach((motive2) => {
      pause_reasons[motive2] = false;
    });
  }
  pause_reasons[motive] = false;
}
function unpaused_call(pause_reasons, fn) {
  let checkPaused;
  return (...args) => {
    clearInterval(checkPaused);
    if (!is_paused(pause_reasons)) {
      fn(...args);
    } else {
      checkPaused = setInterval(() => {
        if (!is_paused(pause_reasons)) {
          fn(...args);
          clearInterval(checkPaused);
        }
      }, 100);
    }
  };
}

// src/frontend/index.ts
class EventEmitter {
  events = {};
  on(event, listener) {
    if (!this.events[event]) {
      this.events[event] = [];
    }
    this.events[event].push(listener);
  }
  off(event, listener) {
    if (!this.events[event])
      return;
    const index = this.events[event].indexOf(listener);
    if (index !== -1) {
      this.events[event].splice(index, 1);
    }
  }
  emit(event, ...args) {
    if (!this.events[event])
      return;
    this.events[event].forEach((listener) => listener(...args));
  }
}
function create_code(name, src) {
  return async () => {
    if (!src || src.length == 0) {
      throw new Error(`missing code: ${name}`);
    }
    if (!src.includes(`
`)) {
      const response = await fetch(src);
      if (!response.ok) {
        throw new Error(`${response.status} code: ${name}`);
      }
      return await response.text();
    }
    return src.replace(/\0/g, "");
  };
}
function create_emiter() {
  return new EventEmitter;
}
async function create_frontend(bus, code, canvas, pause_reasons) {
  const cfg = { init: false };
  if (typeof code.game == "function") {
    code.game = await code.game();
  }
  const bus_emit_resize = unpaused_call(pause_reasons, (width, height) => {
    bus.emit("resize", width, height);
  });
  return {
    native_callback_loop: (dt = 16) => {
      if (!is_paused(pause_reasons)) {
        bus.emit("pad");
        bus.emit("loop", dt);
      }
    },
    native_callback_draw: () => {
      if (!is_paused(pause_reasons)) {
        bus.emit("draw");
      }
    },
    native_callback_init: (width, height) => {
      if (!width || !height) {
        width = canvas.width;
        height = canvas.height;
      }
      if (!is_paused(pause_reasons)) {
        bus.emit("init", width, height, code.game);
      }
      cfg.init = true;
    },
    native_callback_resize: (width, height) => {
      canvas.width = width;
      canvas.height = height;
      if (cfg.init) {
        bus_emit_resize(width, height);
      }
    },
    native_callback_keyboard: (key, value) => {
      if (!is_paused(pause_reasons)) {
        bus.emit("keyboard", key, value);
      }
    }
  };
}

// src/frontend/engine.ts
async function create_engine(hv, canvas, ctx) {
  const methods = () => ({
    frontend: hv.frontend,
    backend: hv.backend,
    pause: (motive) => {
      pause(hv.pause_reasons, motive);
      return methods();
    },
    resume: (motive) => {
      resume(hv.pause_reasons, motive);
      return methods();
    },
    paused: () => {
      return is_paused(hv.pause_reasons);
    },
    getImageData: () => {
      return ctx.getImageData(0, 0, canvas.width, canvas.height);
    },
    stroke: (size) => {
      ctx.lineWidth = size;
      return methods();
    },
    game: (game, not_restart = false) => {
      const type = typeof game;
      if (!["string", "object"].includes(type)) {
        throw new Error(`invalid game format: ${type}`);
      }
      if (not_restart || type === "object") {
        hv.code.game = game;
      } else {
        create_code("game.lua", game)().then((game2) => hv.code.game = game2).then(() => methods().resume("").frontend.native_callback_init());
      }
      return methods();
    },
    on: (key, func) => {
      hv.frontbus.on(key, func);
      return methods();
    }
  });
  return methods();
}

// src/backend/canvas.ts
function hexToStr(n) {
  return (n >>> 0).toString(16).padStart(8, "0");
}
function native_draw_start(render) {
  render.ctx.clearRect(0, 0, render.canvas.width, render.canvas.height);
}
function native_draw_flush(render) {}
function native_draw_clear(render, color, x, y, w, h) {
  render.ctx.fillStyle = "#" + hexToStr(color);
  render.ctx.fillRect(x ?? 0, y ?? 0, w ?? render.canvas.width, h ?? render.canvas.height);
}
function native_draw_color(render, color) {
  const fillstyle = "#" + hexToStr(color);
  render.ctx.strokeStyle = fillstyle;
  render.ctx.fillStyle = fillstyle;
}
function native_draw_line(render, x1, y1, x2, y2) {
  render.ctx.beginPath();
  render.ctx.moveTo(x1, y1);
  render.ctx.lineTo(x2, y2);
  render.ctx.stroke();
}
function native_draw_rect(render, mode, x, y, w, h) {
  if (mode == 1) {
    render.ctx.strokeRect(x, y, w, h);
  } else {
    render.ctx.fillRect(x, y, w, h);
  }
}
function native_draw_poly2(render, mode, verts, x, y, scale, angle, ox, oy) {
  let index = 0;
  render.ctx.beginPath();
  while (index < verts.length) {
    const px = verts[index];
    const py = verts[index + 1];
    const xx = x + (ox - px) * -scale * Math.cos(angle) - (oy - py) * -scale * Math.sin(angle);
    const yy = y + (oy - px) * -scale * Math.sin(angle) + (ox - py) * -scale * Math.cos(angle);
    if (index < 2) {
      render.ctx.moveTo(xx, yy);
    } else {
      render.ctx.lineTo(xx, yy);
    }
    index += 2;
  }
  [
    () => render.ctx.fill(),
    () => {
      render.ctx.closePath();
      render.ctx.stroke();
    },
    () => render.ctx.stroke()
  ][mode]();
}

// src/backend/http.ts
function native_http_handler(self) {
  const method = self.method;
  const headers = new Headers(self.header_dict);
  const params = new URLSearchParams(self.param_dict);
  const url = params.toString() ? `${self.url}?${params.toString()}` : self.url;
  const body = ["HEAD", "GET"].includes(method) ? null : self.body_content;
  self.promise();
  fetch(url, {
    body,
    method,
    headers
  }).then((response) => {
    self.set("ok", response.ok);
    self.set("status", response.status);
    return response.text();
  }).then((content) => {
    self.set("body", content);
    self.resolve();
  }).catch((error) => {
    self.set("ok", false);
    self.set("error", `${error}`);
    self.resolve();
  });
}

// src/backend/image.ts
function preload_image(src) {
  return document.querySelector(`img[src="${src}"]`);
}
async function load_image(src) {
  return new Promise((resolve, reject) => {
    const error = new Error(`img ${src} error!`);
    const el = document.createElement("img");
    el.src = src;
    el.onload = () => resolve(el), el.onerror = () => reject(error);
    el.onabort = () => reject(error);
  });
}
function native_image_load(render, cache, src) {
  if (!(src in cache)) {
    cache[src] = preload_image(src);
    if (!cache[src]) {
      load_image(src).then((el) => cache[src] = el);
    }
  }
}
function native_image_draw(render, cache, src, x, y) {
  native_image_load(render, cache, src);
  if (cache[src]) {
    render.ctx.drawImage(cache[src], x, y);
  }
}

// src/backend/media.ts
function native_media_bootstrap(media, mediatype) {
  const has_support = media.players.map((player) => player.can(mediatype, "", 1)).some((score) => score !== 0);
  media.devices.push(mediatype);
  media.mixer[media.devices.length - 1] = null;
  return has_support ? 1 : 0;
}
function native_media_source(media, channel, url) {
  const type = media.devices[channel];
  const players = media.players.length;
  const best = media.players.map((player, index) => {
    const is_current = player == media.current[channel];
    const score = player.can(type, url, is_current ? 20 : players - index);
    return { player, score };
  }).reduce((max, current) => {
    return current.score > max.score ? current : max;
  });
  if (media.mixer[channel] && (best.score <= 10 || best.player !== media.current[channel])) {
    media.mixer[channel].destroy();
    media.mixer[channel] = null;
  }
  if (!media.mixer[channel] && best.score > 10) {
    media.mixer[channel] = best.player.init(type, channel);
    media.current[channel] = best.player;
  }
  if (!media.mixer[channel]) {
    console.error(`unsupported media: channel ${channel} type ${type}
${url}`);
  }
  media.mixer[channel]?.source(url);
}
function native_media_position(media, channel, x, y) {
  media.mixer[channel]?.position(x, y);
}
function native_media_resize(media, channel, width, height) {
  media.mixer[channel]?.resize(width, height);
}
function native_media_play(media, channel) {
  media.mixer[channel]?.play();
}
function native_media_resume(media, channel) {
  media.mixer[channel]?.resume();
}
function native_media_pause(media, channel) {
  media.mixer[channel]?.pause();
}
function native_media_stop(media, channel) {
  media.mixer[channel]?.destroy();
  media.mixer[channel] = null;
}
function native_media_time(media, channel, time) {
  media.mixer[channel]?.set_time(time);
}

// src/backend/text.ts
function font_apply_if_changes(render, font) {
  if (font.name != font.old.name || font.size != font.old.size) {
    render.ctx.font = `${font.size}px ${font.name}`;
    render.ctx.textBaseline = "top";
    render.ctx.textAlign = "left";
  }
}
function native_text_font_name(render, font, name) {
  font.name = name;
}
function native_text_font_size(render, font, size) {
  font.size = Math.floor(size);
}
function native_text_font_default(render, font, id) {
  font.name = "sans";
}
function native_text_font_previous(render, font) {
  const [name, size] = [font.name, font.size];
  font.name = font.old.name;
  font.size = font.old.size;
  font.old.name = name;
  font.old.size = size;
}
function native_text_print(render, font, x, y, text) {
  font_apply_if_changes(render, font);
  render.ctx.fillText(text, x, y);
}
function native_text_mensure(render, font, text) {
  font_apply_if_changes(render, font);
  const { width, actualBoundingBoxAscent, actualBoundingBoxDescent } = render.ctx.measureText(text);
  return [width, actualBoundingBoxAscent + actualBoundingBoxDescent];
}

// src/backend/system.ts
function native_system_get_language() {
  return navigator.language;
}

// src/backend/index.ts
function create_canvas(canvas) {
  if (typeof canvas == "object") {
    return canvas;
  }
  if (typeof canvas == "string") {
    const el = document.querySelector(canvas);
    if (!el) {
      throw new Error(`element dont exist: ${canvas}`);
    }
    return el;
  }
  return document.createElement("canvas");
}
function create_backend(canvas, ctx, players) {
  const render = { canvas, ctx };
  const text_cache = { name: "sans", size: 5, old: { name: "sans", size: 8 } };
  const image_cache = {};
  const media_cache = { devices: [], current: [], mixer: {}, players };
  media_cache.players.push({ can: () => 0 });
  return {
    native_http_handler: (self) => native_http_handler(self),
    native_draw_start: () => native_draw_start(render),
    native_draw_flush: () => native_draw_flush(render),
    native_draw_color: (color) => native_draw_color(render, color),
    native_draw_clear: (color, x, y, w, h) => native_draw_clear(render, color, x, y, w, h),
    native_draw_rect: (mode, x, y, w, h) => native_draw_rect(render, mode, x, y, w, h),
    native_draw_line: (x1, y1, x2, y2) => native_draw_line(render, x1, y1, x2, y2),
    native_draw_poly2: (mode, verts, x, y, scale, angle, ox, oy) => native_draw_poly2(render, mode, verts, x, y, scale, angle, ox, oy),
    native_text_font_name: (name) => native_text_font_name(render, text_cache, name),
    native_text_font_default: (id) => native_text_font_default(render, text_cache, id),
    native_text_font_size: (size) => native_text_font_size(render, text_cache, size),
    native_text_font_previous: (size) => native_text_font_previous(render, text_cache),
    native_text_print: (x, y, text) => native_text_print(render, text_cache, x, y, text),
    native_text_mensure: (text) => native_text_mensure(render, text_cache, text),
    native_image_load: (src) => native_image_load(render, image_cache, src),
    native_image_draw: (src, x, y) => native_image_draw(render, image_cache, src, x, y),
    native_system_get_language: () => native_system_get_language(),
    native_media_bootstrap: (mediatype) => native_media_bootstrap(media_cache, mediatype),
    native_media_source: (channel, url) => native_media_source(media_cache, channel, url),
    native_media_position: (channel, x, y) => native_media_position(media_cache, channel, x, y),
    native_media_resize: (channel, width, height) => native_media_resize(media_cache, channel, width, height),
    native_media_resume: (channel) => native_media_resume(media_cache, channel),
    native_media_play: (channel) => native_media_play(media_cache, channel),
    native_media_stop: (channel) => native_media_stop(media_cache, channel),
    native_media_pause: (channel) => native_media_pause(media_cache, channel),
    native_media_time: (channel, time) => native_media_time(media_cache, channel, time),
    native_draw_poly: (mode, verts, x = 0, y = 0, scale = 1, angle = 0, ox = 0, oy = 0) => native_draw_poly2(render, mode, verts, x, y, scale, angle, ox, oy),
    native_draw_image: (src, x, y) => native_image_draw(render, image_cache, src, x, y),
    native_draw_text: (x, y, text) => {
      typeof x == "number" && (text || text == 0) && native_text_print(render, text_cache, x, y, `${text}`);
      return native_text_mensure(render, text_cache, `${text ?? x}`);
    },
    native_draw_font: (name, size) => {
      native_text_font_name(render, text_cache, name);
      native_text_font_size(render, text_cache, size);
    }
  };
}

// src/driver/resize.ts
async function prepare(hv) {}
async function install(hv) {}
async function startup(hv, cfg = { widescreen: true }) {
  const set_size = (device) => {
    const width = device.innerWidth;
    const height = device.innerHeight;
    const widescreen = cfg.widescreen && height >= width;
    hv.frontend.native_callback_resize(width, widescreen ? height / 2 : height);
  };
  window.addEventListener("resize", (ev) => set_size(ev.target));
  set_size(window);
}
var resize_default = {
  prepare,
  install,
  startup
};

// src/driver/runtime.ts
async function prepare2(hv) {}
async function install2(hv) {}
async function startup2(hv, cfg = { uptime: false, unfocus_pause: false }) {
  let uptime = performance.now();
  function tick() {
    let new_time = performance.now();
    const dt = cfg.uptime ? new_time : new_time - uptime;
    uptime = new_time;
    hv.frontend.native_callback_loop(dt);
    hv.frontend.native_callback_draw();
    window.requestAnimationFrame(tick);
  }
  if (cfg.unfocus_pause) {
    window.addEventListener("blur", () => pause(hv.pause_reasons, "focus"));
    window.addEventListener("focus", () => resume(hv.pause_reasons, "focus"));
  }
  await (() => new Promise((r) => setTimeout(r, 1)))();
  window.requestAnimationFrame(() => {
    hv.frontend.native_callback_init();
    tick();
  });
}
var runtime_default = {
  prepare: prepare2,
  install: install2,
  startup: startup2
};

// presets.json
var padmap = {
  default: {
    pads: ["a", "b", "c", "d", "a", "b", "c", "d", "menu", "menu", "menu", "menu", "up", "down", "left", "right"],
    axis: ["left", "right", "up", "down", "left", "right", "up", "down", "left", "right", "up", "down"]
  },
  legacy: {
    pads: ["red", "green", "yellow", "blue", "left", "right", "up", "down", "menu"],
    axis: ["left", "right", "up", "down", "left", "right", "up", "down"]
  }
};
var keymap = {
  default: [
    [13, "a"],
    [38, "up"],
    [37, "left"],
    [40, "down"],
    [39, "right"],
    [403, "a"],
    [404, "b"],
    [405, "c"],
    [406, "d"],
    [10009, "menu"],
    ["KeyZ", "a"],
    ["KeyX", "b"],
    ["KeyC", "c"],
    ["KeyV", "d"],
    ["Enter", "a"],
    ["ArrowUp", "up"],
    ["ArrowDown", "down"],
    ["ArrowLeft", "left"],
    ["ArrowRight", "right"],
    ["ShiftLeft", "menu"]
  ],
  legacy: [
    [13, "enter"],
    [38, "up"],
    [37, "left"],
    [40, "down"],
    [39, "right"],
    [403, "red"],
    [404, "green"],
    [405, "yellow"],
    [406, "blue"],
    [10009, "enter"],
    ["KeyZ", "red"],
    ["KeyX", "green"],
    ["KeyC", "yellow"],
    ["KeyV", "blue"],
    ["Enter", "enter"],
    ["ArrowUp", "up"],
    ["ArrowDown", "down"],
    ["ArrowLeft", "left"],
    ["ArrowRight", "right"]
  ]
};

// src/frontend/gamepad.ts
function gamepad_trigger(trigger, gamepadmap) {
  const previousStates = new Array(gamepadmap.pads.length).fill(0);
  const previousAxisStates = new Array(gamepadmap.axis.length).fill(0);
  const deadZone = 0.2;
  return () => {
    const gamepads = navigator?.getGamepads() || [];
    for (const gamepad of gamepads) {
      if (gamepad) {
        gamepadmap.pads.forEach((action, index) => {
          const isPressed = gamepad.buttons[index]?.pressed ? 1 : 0;
          if (isPressed !== previousStates[index]) {
            trigger(action, isPressed);
            previousStates[index] = isPressed;
          }
        });
        gamepadmap.axis.forEach((action, index) => {
          let isPressed = 0;
          let axisValue = gamepad.axes[Math.floor(index / 2)];
          if (index & 1 && deadZone < axisValue) {
            isPressed = 1;
          }
          if (!(index & 1) && -deadZone > axisValue) {
            isPressed = 1;
          }
          if (isPressed !== previousAxisStates[index]) {
            trigger(action, isPressed);
            previousAxisStates[index] = isPressed;
          }
        });
      }
    }
  };
}

// src/driver/gamepad.ts
async function prepare3(hv) {}
async function install3(hv) {}
async function startup3(hv, keys) {
  if (typeof keys != "object") {
    keys = `${keys}` in padmap ? padmap[keys] : padmap["default"];
  }
  hv.frontbus.on("pad", gamepad_trigger((key, value) => hv.frontbus.emit("keyboard", key, value), keys));
}
var gamepad_default = {
  prepare: prepare3,
  install: install3,
  startup: startup3
};

// src/frontend/keyboard.ts
function keyboard_trigger(trigger, keymap2) {
  return (ev) => {
    const key = keymap2.find((key2) => [ev.code, ev.keyCode].includes(key2[0]));
    if (key) {
      ev.preventDefault();
      trigger(key[1], Number(ev.type === "keydown"));
    }
  };
}

// src/driver/keyboard.ts
async function prepare4(hv) {}
async function install4(hv) {}
async function startup4(hv, keys) {
  if (typeof keys != "object") {
    keys = `${keys}` in keymap ? keymap[keys] : keymap["default"];
  }
  document.addEventListener("keydown", keyboard_trigger(hv.frontend.native_callback_keyboard, keys));
  document.addEventListener("keyup", keyboard_trigger(hv.frontend.native_callback_keyboard, keys));
}
var keyboard_default = {
  prepare: prepare4,
  install: install4,
  startup: startup4
};

// src/driver/lua/check.ts
var mock = async (_) => {};
async function check_wasmoon(hv) {
  if (!hv.vm.lua) {
    throw new Error("wasmoon is required!");
  }
}
async function check_fengari(hv) {
  if (!hv.vm.fengari) {
    throw new Error("fengari is required!");
  }
}
var check_default = {
  wasmoon: {
    prepare: mock,
    install: check_wasmoon,
    startup: mock
  },
  fengari: {
    prepare: mock,
    install: check_fengari,
    startup: mock
  },
  fengari_wasmoon: {
    install: async (hv) => {
      let errors = 0;
      await Promise.allSettled([check_wasmoon(hv), check_fengari(hv)]).then((results) => {
        errors = results.filter((result) => result.status === "rejected").length;
      });
      if (errors >= 2) {
        throw new Error("wamoon or fengari is required!");
      }
    },
    prepare: mock,
    startup: mock
  }
};

// src/driver/lua/wasmoon.ts
async function prepare5(hv, LuaFactory) {
  if (!LuaFactory || hv.vm.fengari) {
    return;
  }
  hv.vm.factory = new LuaFactory;
  hv.vm.lua = await hv.vm.factory.createEngine();
}
async function install5(hv, _, LuaMultiReturn) {
  if (!hv.vm.lua) {
    return;
  }
  for (const key in hv.backend) {
    hv.vm.lua.global.set(key, hv.backend[key]);
  }
  hv.vm.lua.global.set("native_text_mensure", (text) => {
    return LuaMultiReturn.from(hv.backend.native_text_mensure(text));
  });
  hv.vm.lua.global.set("native_draw_text", (x, y, text) => {
    return LuaMultiReturn.from(hv.backend.native_draw_text(x, y, text));
  });
  hv.vm.lua.global.set("native_dict_poly", {
    poly: hv.backend.native_draw_poly,
    poly2: hv.backend.native_draw_poly
  });
  hv.vm.lua.global.set("native_http_force_protocol", window.location.protocol == "http:" ? "http" : "https");
  hv.vm.lua.global.set("native_http_has_ssl", true);
  hv.vm.lua.global.set("native_json_encode", JSON.stringify);
  hv.vm.lua.global.set("native_json_decode", JSON.parse);
  hv.vm.lua.global.set("native_base64_encode", atob);
  hv.vm.lua.global.set("native_base64_decode", btoa);
  await hv.vm.lua.doString(await hv.code.engine());
  for (const key in hv.frontend) {
    hv.frontbus.on(key.replace(/^native_callback_/, ""), hv.vm.lua.global.get(key));
  }
}
async function startup5(hv) {}
var wasmoon_default = {
  prepare: prepare5,
  install: install5,
  startup: startup5
};

// src/driver/lua/fengari.ts
async function prepare_jsonrxi(hv, json_lib_rxi) {
  if (!hv.vm.fengari) {
    return;
  }
  const pseudonym = "jsonrxi.lua";
  const lua_lib = await create_code(pseudonym, json_lib_rxi)();
  const lua_code = lua_lib.replace("json.encode", "native_json_encode").replace("json.decode", "native_json_decode");
  const lua_buffer = hv.vm.fengari.to_luastring(lua_code);
  hv.vm.fengari.lauxlib.luaL_loadbuffer(hv.vm.fengari.L, lua_buffer, lua_code.length, pseudonym);
  hv.vm.fengari.lua.lua_pcall(hv.vm.fengari.L, 0, 0, 0);
}
async function prepare6(hv, fengari) {
  if (!fengari || hv.vm.lua) {
    return;
  }
  fengari.lualib.luaL_openlibs(fengari.L);
  hv.vm.fengari = fengari;
}
async function install6(hv, fengari) {
  if (!hv.vm.fengari) {
    return;
  }
  const httplua = (reqid, key, data) => {
    const params = data !== undefined ? 3 : 2;
    fengari.lua.lua_getglobal(fengari.L, fengari.to_luastring("native_callback_http"));
    fengari.lua.lua_pushinteger(fengari.L, reqid);
    fengari.lua.lua_pushstring(fengari.L, fengari.to_luastring(key));
    if (typeof data == "string") {
      fengari.lua.lua_pushstring(fengari.L, fengari.to_luastring(data));
    }
    if (typeof data == "number") {
      fengari.lua.lua_pushnumber(fengari.L, data);
    }
    if (typeof data == "boolean") {
      fengari.lua.lua_pushboolean(fengari.L, data);
    }
    if (fengari.lua.lua_pcall(fengari.L, params, 1, 0) !== 0) {
      throw fengari.to_jsstring(fengari.lua.lua_tostring(fengari.L, -1));
    }
    if (fengari.lua.lua_type(fengari.L, -1) == fengari.lua.LUA_TSTRING) {
      return fengari.to_jsstring(fengari.lua.lua_tostring(fengari.L, -1));
    }
  };
  const define_lua_callback = (func_name, func_decorator) => {
    hv.frontbus.on(func_name.replace(/^native_callback_/, ""), (a, b, c, d, e, f) => {
      fengari.lua.lua_getglobal(fengari.L, fengari.to_luastring(func_name));
      const func = func_decorator ?? (() => 0);
      const res = func(a, b, c, d, e, f) ?? 0;
      if (fengari.lua.lua_pcall(fengari.L, res, 0, 0) !== 0) {
        const error_message = fengari.lua.lua_tostring(fengari.L, -1);
        throw error_message && fengari.to_jsstring(error_message) || func_name;
      }
    });
  };
  const define_lua_func = (func_name, func_decorator) => {
    const func_native = hv.backend[func_name];
    fengari.lua.lua_pushcfunction(fengari.L, () => {
      try {
        const res = func_decorator(func_native);
        return res ?? 0;
      } catch (e) {
        throw `${e} in ${func_name}`;
      }
    });
    fengari.lua.lua_setglobal(fengari.L, fengari.to_luastring(func_name));
  };
  define_lua_func("native_draw_start", (func) => {
    func();
  });
  define_lua_func("native_draw_flush", (func) => {
    func();
  });
  define_lua_func("native_draw_clear", (func) => {
    const color = fengari.lua.lua_tointeger(fengari.L, 1) >>> 0;
    const x = fengari.lua.lua_tonumber(fengari.L, 2);
    const y = fengari.lua.lua_tonumber(fengari.L, 3);
    const w = fengari.lua.lua_tonumber(fengari.L, 4);
    const h = fengari.lua.lua_tonumber(fengari.L, 5);
    func(color, x, y, w, h);
  });
  define_lua_func("native_draw_color", (func) => {
    const color = fengari.lua.lua_tointeger(fengari.L, 1) >>> 0;
    func(color);
  });
  define_lua_func("native_draw_rect", (func) => {
    const mode = fengari.lua.lua_tointeger(fengari.L, 1);
    const x = fengari.lua.lua_tonumber(fengari.L, 2);
    const y = fengari.lua.lua_tonumber(fengari.L, 3);
    const w = fengari.lua.lua_tonumber(fengari.L, 4);
    const h = fengari.lua.lua_tonumber(fengari.L, 5);
    func(mode, x, y, w, h);
  });
  define_lua_func("native_draw_line", (func) => {
    const x1 = fengari.lua.lua_tonumber(fengari.L, 1);
    const y1 = fengari.lua.lua_tonumber(fengari.L, 2);
    const x2 = fengari.lua.lua_tonumber(fengari.L, 3);
    const y2 = fengari.lua.lua_tonumber(fengari.L, 4);
    func(x1, y1, x2, y2);
  });
  define_lua_func("native_draw_poly2", (func) => {
    let i = 1;
    const mode = fengari.lua.lua_tointeger(fengari.L, 1);
    const verts = [];
    const x = fengari.lua.lua_tonumber(fengari.L, 3);
    const y = fengari.lua.lua_tonumber(fengari.L, 4);
    const scale = fengari.lua.lua_tonumber(fengari.L, 5);
    const angle = fengari.lua.lua_tonumber(fengari.L, 6);
    const ox = fengari.lua.lua_tonumber(fengari.L, 7);
    const oy = fengari.lua.lua_tonumber(fengari.L, 8);
    while (fengari.lua.lua_rawgeti(fengari.L, 2, i) !== fengari.lua.LUA_TNIL) {
      verts.push(fengari.lua.lua_tonumber(fengari.L, -1));
      fengari.lua.lua_pop(fengari.L, 1);
      i++;
    }
    func(mode, verts, x, y, scale, angle, ox, oy);
  });
  define_lua_func("native_text_print", (func) => {
    const x = fengari.lua.lua_tonumber(fengari.L, 1);
    const y = fengari.lua.lua_tonumber(fengari.L, 2);
    const text = fengari.to_jsstring(fengari.lua.lua_tostring(fengari.L, 3));
    func(x, y, text);
  });
  define_lua_func("native_text_font_size", (func) => {
    const size = fengari.lua.lua_tonumber(fengari.L, 1);
    func(size);
  });
  define_lua_func("native_text_font_name", (func) => {
    const name = fengari.to_jsstring(fengari.lua.lua_tostring(fengari.L, 1));
    func(name);
  });
  define_lua_func("native_text_font_default", (func) => {
    func();
  });
  define_lua_func("native_text_font_previous", (func) => {
    func();
  });
  define_lua_func("native_text_mensure", (func) => {
    const text = fengari.to_jsstring(fengari.lua.lua_tostring(fengari.L, 1));
    const [width, height] = func(text);
    fengari.lua.lua_pushnumber(fengari.L, width);
    fengari.lua.lua_pushnumber(fengari.L, height);
    return 2;
  });
  define_lua_func("native_system_get_language", (func) => {
    fengari.lua.lua_pushstring(fengari.L, func());
    return 1;
  });
  define_lua_func("native_image_load", (func) => {
    const src = fengari.to_jsstring(fengari.lua.lua_tostring(fengari.L, 1));
    func(src);
  });
  define_lua_func("native_image_draw", (func) => {
    const src = fengari.to_jsstring(fengari.lua.lua_tostring(fengari.L, 1));
    const x = fengari.lua.lua_tonumber(fengari.L, 2);
    const y = fengari.lua.lua_tonumber(fengari.L, 3);
    func(src, x, y);
  });
  define_lua_func("native_media_bootstrap", (func) => {
    const mediatype = fengari.to_jsstring(fengari.lua.lua_tostring(fengari.L, 1));
    const channels = func(mediatype);
    fengari.lua.lua_pushnumber(fengari.L, channels);
    return 1;
  });
  define_lua_func("native_media_source", (func) => {
    const channel = fengari.lua.lua_tonumber(fengari.L, 1);
    const src = fengari.to_jsstring(fengari.lua.lua_tostring(fengari.L, 2));
    func(channel, src);
  });
  define_lua_func("native_media_position", (func) => {
    const channel = fengari.lua.lua_tonumber(fengari.L, 1);
    const x = fengari.lua.lua_tonumber(fengari.L, 2);
    const y = fengari.lua.lua_tonumber(fengari.L, 3);
    func(channel, x, y);
  });
  define_lua_func("native_media_resize", (func) => {
    const channel = fengari.lua.lua_tonumber(fengari.L, 1);
    const w = fengari.lua.lua_tonumber(fengari.L, 2);
    const h = fengari.lua.lua_tonumber(fengari.L, 3);
    func(channel, w, h);
  });
  define_lua_func("native_media_time", (func) => {
    const channel = fengari.lua.lua_tonumber(fengari.L, 1);
    const time = fengari.lua.lua_tonumber(fengari.L, 2);
    func(channel, Math.floor(time));
  });
  define_lua_func("native_media_resume", (func) => {
    func(fengari.lua.lua_tonumber(fengari.L, 1));
  });
  define_lua_func("native_media_play", (func) => {
    func(fengari.lua.lua_tonumber(fengari.L, 1));
  });
  define_lua_func("native_media_pause", (func) => {
    func(fengari.lua.lua_tonumber(fengari.L, 1));
  });
  define_lua_func("native_media_stop", (func) => {
    func(fengari.lua.lua_tonumber(fengari.L, 1));
  });
  define_lua_func("native_http_handler", (func) => {
    const request_id = fengari.lua.lua_tonumber(fengari.L, 2);
    func({
      param_dict: {},
      header_dict: {},
      set: (key, value) => httplua(request_id, `set-${key}`, value),
      promise: () => httplua(request_id, "async-promise"),
      resolve: () => httplua(request_id, "async-resolve"),
      method: httplua(request_id, "get-method"),
      body: httplua(request_id, "get-body"),
      url: httplua(request_id, "get-fullurl")
    });
  });
  fengari.lua.lua_pushboolean(fengari.L, true);
  fengari.lua.lua_setglobal(fengari.L, fengari.to_luastring("native_http_has_callback"));
  fengari.lua.lua_pushboolean(fengari.L, true);
  fengari.lua.lua_setglobal(fengari.L, fengari.to_luastring("native_http_has_ssl"));
  fengari.lua.lua_pushstring(fengari.L, fengari.to_luastring(window.location.protocol == "http:" ? "http" : "https"));
  fengari.lua.lua_setglobal(fengari.L, fengari.to_luastring("native_http_force_protocol"));
  const lua_engine = await hv.code.engine();
  fengari.lauxlib.luaL_loadbuffer(fengari.L, fengari.to_luastring(lua_engine), lua_engine.length, "engine");
  fengari.lua.lua_pcall(fengari.L, 0, 0, 0);
  define_lua_callback("native_callback_init", (width, height, game) => {
    fengari.lua.lua_pushnumber(fengari.L, width);
    fengari.lua.lua_pushnumber(fengari.L, height);
    fengari.lauxlib.luaL_loadbuffer(fengari.L, fengari.to_luastring(game), game.lenght, "game");
    fengari.lua.lua_pcall(fengari.L, 0, 1, 0);
    return 3;
  });
  define_lua_callback("native_callback_draw");
  define_lua_callback("native_callback_loop", (dt) => {
    fengari.lua.lua_pushnumber(fengari.L, dt);
    return 1;
  });
  define_lua_callback("native_callback_resize", (width, height) => {
    fengari.lua.lua_pushnumber(fengari.L, width);
    fengari.lua.lua_pushnumber(fengari.L, height);
    return 2;
  });
  define_lua_callback("native_callback_keyboard", (key, value) => {
    fengari.lua.lua_pushstring(fengari.L, fengari.to_luastring(key));
    fengari.lua.lua_pushboolean(fengari.L, value);
    return 2;
  });
}
async function startup6(hv, fengari) {}
var fengari_default = {
  jsonrxi: {
    prepare: prepare_jsonrxi,
    install: async (hv) => {},
    startup: async (hv) => {}
  },
  prepare: prepare6,
  install: install6,
  startup: startup6
};

// src/driver/players/fake.ts
function set_time(time) {
  return Math.max(0, performance.now() - time);
}
function genVerts(ms, width, height) {
  const pts = [];
  const duration = 1000;
  const points = [
    [width / 2, 0],
    [width, 0],
    [width, height],
    [0, height],
    [0, 0],
    [width / 2, 0]
  ];
  const segmentTime = duration / (points.length - 1);
  const step = Math.min(Math.floor(ms / segmentTime), points.length - 2);
  const t = ms % segmentTime / segmentTime;
  function getLastPoint() {
    const [x1, y1] = points[step];
    const [x2, y2] = points[step + 1];
    return [x1 + (x2 - x1) * t, y1 + (y2 - y1) * t];
  }
  pts.push(width / 2, height / 2);
  points.filter((_, index) => index < step + 1).forEach((points2) => pts.push(points2[0], points2[1]));
  const [x, y] = getLastPoint();
  pts.push(x, y);
  return pts;
}
function cronomether(ms) {
  let date = new Date(ms);
  let formattedTime = date.toISOString().substr(11, 8);
  let milliseconds = (ms % 1000).toFixed(0).padStart(3, "0");
  return `${formattedTime}:${milliseconds}`;
}
function can(type, url, score, default_score) {
  if (typeof default_score !== "number") {
    default_score = 11;
  }
  if (!["video", "tv", "youtube", "stream"].includes(type)) {
    return 0;
  }
  return score + default_score;
}
function init(type, channel) {
  const el_root = document.querySelector("main");
  const el_media = document.createElement("canvas");
  const ctx = el_media.getContext("2d");
  const render = { canvas: el_media, ctx };
  let src = "";
  let cmd = "init";
  let paused = true;
  let started = performance.now();
  el_media.width = el_root.clientWidth;
  el_media.height = el_root.clientHeight;
  el_media.style.zIndex = `${-channel - 1}`;
  el_root.appendChild(el_media);
  function tick() {
    if (!el_media.parentElement) {
      return;
    }
    const uptime = performance.now() - started;
    const crono = cronomether(uptime);
    const [cor1, cor2] = [255, 572662527];
    const font1 = (el_media.width / 80).toFixed(0);
    const font2 = (el_media.width / 16).toFixed(0);
    const secs = Math.ceil(uptime / 1000);
    native_draw_clear(render, secs & 1 ? cor1 : cor2);
    native_draw_color(render, secs & 1 ? cor2 : cor1);
    native_draw_poly2(render, 0, genVerts(uptime % 1000, el_media.width, el_media.height), 0, 0, 1, 0, 0, 0);
    native_draw_color(render, 858993663);
    native_draw_line(render, 0, el_media.height / 2, el_media.width, el_media.height / 2);
    native_draw_line(render, el_media.width / 2, 0, el_media.width / 2, el_media.height);
    native_draw_color(render, 4294967295);
    ctx.font = `${font1}px sans`;
    ctx.textBaseline = "top";
    ctx.textAlign = "left";
    ctx.fillText(type, 0, 0);
    ctx.textBaseline = "bottom";
    ctx.fillText(cmd, 0, el_media.height);
    ctx.textAlign = "right";
    ctx.fillText(src, el_media.width, el_media.height);
    ctx.textBaseline = "top";
    ctx.fillText(crono, el_media.width, 0);
    ctx.font = `${font2}px sans`;
    ctx.textBaseline = "middle";
    ctx.textAlign = "center";
    ctx.fillText(`${secs}`, el_media.width / 2, el_media.height / 2);
    if (!paused) {
      window.requestAnimationFrame(tick);
    }
  }
  return {
    can,
    set_time: (time) => {
      started = set_time(time);
    },
    source: (url) => {
      cmd = "source";
      src = url;
    },
    play: () => {
      cmd = "play";
      started = set_time(0);
      if (paused) {
        paused = false;
        tick();
      }
    },
    pause: () => {
      cmd = "pause";
      paused = true;
    },
    resume: () => {
      cmd = "resume";
      if (paused) {
        paused = false;
        tick();
      }
    },
    resize: (width, height) => {
      cmd = "resize";
      el_media.width = width;
      el_media.height = height;
    },
    position: (x, y) => {
      cmd = "position";
      el_media.style.left = `${x}px`;
      el_media.style.top = `${y}px`;
    },
    destroy: () => {
      el_media.remove();
    }
  };
}
async function create_player(hv, default_score) {
  hv.media_players.push({ init, can: (a, b, c) => can(a, b, c, default_score) });
}
var fake_default = {
  prepare: create_player,
  install: async () => {},
  startup: async () => {}
};

// src/driver/players/html5.ts
var is_video = (type, func) => type === "video" ? func : () => {};
function set_time2(el) {
  return function(time) {
    if (el.fastSeek) {
      el.fastSeek(time);
    } else {
      el.currentTime = time;
    }
  };
}
function can2(type, url, score) {
  if (!["video", "audio", "music", "sfx"].includes(type)) {
    return 0;
  }
  return score + 20;
}
function init2(type, channel) {
  let el_media = document.createElement(type === "video" ? "video" : "audio");
  const el_root = document.querySelector("main");
  is_video(type, () => el_media.style.zIndex = `${-channel - 1}`)();
  el_root.appendChild(el_media);
  return {
    can: can2,
    set_time: set_time2(el_media),
    source: (url) => {
      el_media.src = url;
    },
    play: () => {
      set_time2(el_media)(0);
      el_media.play();
    },
    pause: () => {
      el_media.pause();
    },
    resume: () => {
      el_media.play();
    },
    resize: is_video(type, (width, height) => {
      el_media.style.width = `${width}px`;
      el_media.style.height = `${height}px`;
    }),
    position: is_video(type, (x, y) => {
      el_media.style.left = `${x}px`;
      el_media.style.top = `${y}px`;
    }),
    destroy: () => {
      const drop = () => el_media.remove();
      el_media.onerror = drop;
      el_media.onabort = drop;
      el_media.onload = drop;
      el_media.src = "";
      el_media.load();
    }
  };
}
async function create_player2(hv) {
  hv.media_players.push({ init: init2, can: can2 });
}
var html5_default = {
  prepare: create_player2,
  install: async () => {},
  startup: async () => {}
};

// src/driver/players/videojs.ts
function can3(type, url, score) {
  if (!["video", "stream"].includes(type)) {
    return 0;
  }
  if (/\.(m3u8|mpd)$/i.test(url)) {
    return score + 40;
  }
  return score + 20;
}
function init3(videojslib, type, channel) {
  const el_root = document.querySelector("main");
  const el_media = document.createElement("video");
  el_media.className = "video-js";
  el_root.appendChild(el_media);
  if (!document.querySelector("#vsj-hidden")) {
    const style = document.createElement("style");
    style.innerHTML = ".vsj-hidden, .video-js div, .video-js button { display: none }";
    style.id = "vsj-hidden";
    document.head.appendChild(style);
  }
  el_media.style.zIndex = `${-channel - 1}`;
  const player = videojslib(el_media, {
    controls: false
  });
  return {
    can: can3,
    set_time: (time) => {
      player.currentTime(time);
    },
    source: (url) => {
      player.src({ src: url });
    },
    play: () => {
      player.currentTime(0);
      player.play();
    },
    pause: () => {
      player.pause();
    },
    resume: () => {
      player.play();
    },
    resize: (width, height) => {
      player.width(width);
      player.height(height);
      el_media.style.width = `${width}px`;
      el_media.style.height = `${height}px`;
    },
    position: (x, y) => {
      el_media.style.left = `${x}px`;
      el_media.style.top = `${y}px`;
    },
    destroy: () => {
      player.dispose();
      const drop = () => el_media.remove();
      el_media.onabort = drop;
      el_media.onerror = drop;
      el_media.onload = drop;
      el_media.src = "";
      el_media.load();
      el_media.remove();
    }
  };
}
async function create_player3(hv, videojslib) {
  if (videojslib) {
    hv.media_players.push({ can: can3, init: (a, b) => init3(videojslib, a, b) });
  }
}
var videojs_default = {
  prepare: create_player3,
  install: async () => {},
  startup: async () => {}
};

// src/driver/players/youtube.ts
function youtubeId(url) {
  const regex = /(?:youtu\.be\/|youtube\.com\/(?:watch\?(?:.*&)?v=|embed\/|v\/))([^?&"'>]+)/;
  const match = url.match(regex);
  return match ? match[1] : null;
}
function can4(type, url, score) {
  if (!["video", "youtube"].includes(type)) {
    return 0;
  }
  if (url.length === 0) {
    return 1;
  }
  if (youtubeId(url)) {
    return 100;
  }
  return 0;
}
function init4(type, channel) {
  const el_root = document.querySelector("main");
  const iframe = document.createElement("iframe");
  iframe.className = "youtube-player";
  iframe.frameBorder = "0";
  iframe.allow = "autoplay encrypted-media";
  iframe.allowFullscreen = true;
  el_root.appendChild(iframe);
  let isPlayerReady = false;
  iframe.style.zIndex = `${-channel - 1}`;
  iframe.addEventListener("load", () => {
    isPlayerReady = true;
  });
  function sendCommandWhenReady(func, args = []) {
    if (isPlayerReady) {
      iframe.contentWindow?.postMessage(JSON.stringify({ event: "command", func, args }), "*");
    } else {
      setTimeout(() => sendCommandWhenReady(func, args), 100);
    }
  }
  return {
    can: can4,
    set_time: (time) => {
      sendCommandWhenReady("seekTo", [time, true]);
    },
    source: (url) => {
      iframe.src = `https://www.youtube.com/embed/${youtubeId(url)}?enablejsapi=1&autoplay=0&modestbranding=1&controls=0`;
    },
    play: () => {
      sendCommandWhenReady("playVideo");
    },
    pause: () => {
      sendCommandWhenReady("pauseVideo");
    },
    resume: () => {
      sendCommandWhenReady("playVideo");
    },
    resize: (width, height) => {
      iframe.style.width = `${width}px`;
      iframe.style.height = `${height}px`;
    },
    position: (x, y) => {
      iframe.style.left = `${x}px`;
      iframe.style.top = `${y}px`;
    },
    destroy: () => {
      el_root.removeChild(iframe);
    }
  };
}
async function create_player4(hv) {
  hv.media_players.push({ init: init4, can: can4 });
}
var youtube_default = {
  prepare: create_player4,
  install: async () => {},
  startup: async () => {}
};

// src/driver/index.ts
var none = {
  prepare: async (hv) => {},
  install: async (hv) => {},
  startup: async (hv) => {}
};
var driver_map = {
  none,
  resize: resize_default,
  fengari: fengari_default,
  wasmoon: wasmoon_default,
  runtime: runtime_default,
  gamepad: gamepad_default,
  keyboard: keyboard_default,
  "player-fake": fake_default,
  "player-html5": html5_default,
  "player-videojs": videojs_default,
  "player-youtube": youtube_default,
  "wasmoon-check": check_default.wasmoon,
  "fengari-check": check_default.fengari,
  "fengari-or-wasmoon-check": check_default.fengari_wasmoon,
  "fengari-jsonrxi": fengari_default.jsonrxi
};
function custom_driver(step_name, driver_name) {
  return async (hv, func, ...args) => {
    if (typeof func !== "object" || typeof func[step_name] !== "function") {
      throw new Error(`driver not found: ${driver_name} (${step_name})`);
    }
    await func[step_name](hv, ...args);
  };
}
function get_driver(driver_name) {
  const driver = driver_map[driver_name];
  if (driver) {
    return driver;
  }
  return {
    prepare: custom_driver("prepare", driver_name),
    install: custom_driver("install", driver_name),
    startup: custom_driver("startup", driver_name)
  };
}

// src/index.ts
var src_default = (() => {
  let cfg_game;
  let cfg_engine;
  let cfg_libs = [];
  let cfg_canvas;
  let cfg_rootel;
  const methods = () => ({
    set_game: (game_code) => {
      cfg_game = game_code;
      return methods();
    },
    set_engine: (engine_code) => {
      cfg_engine = engine_code;
      return methods();
    },
    set_el_root: (element) => {
      cfg_rootel = element;
      return methods();
    },
    set_el_canvas: (canvas) => {
      cfg_canvas = canvas;
      return methods();
    },
    set_library: (type, ...args) => {
      cfg_libs.push({
        driver: get_driver(type),
        args
      });
      return methods();
    },
    build: async () => {
      const vm = {};
      const media_players = [];
      const pause_reasons = {};
      const code = {
        game: typeof cfg_game == "string" ? create_code("game.lua", cfg_game) : cfg_game,
        engine: create_code("engine.lua", cfg_engine)
      };
      const canvas = create_canvas(cfg_canvas);
      const ctx = canvas.getContext("2d");
      const backend = create_backend(canvas, ctx, media_players);
      const frontbus = create_emiter();
      const frontend = await create_frontend(frontbus, code, canvas, pause_reasons);
      const hypervisor = {
        vm,
        code,
        backend,
        frontend,
        frontbus,
        pause_reasons,
        media_players
      };
      await Promise.all(cfg_libs.map((lib) => lib.driver.prepare(hypervisor, ...lib.args)));
      await Promise.all(cfg_libs.map((lib) => lib.driver.install(hypervisor, ...lib.args)));
      await Promise.all(cfg_libs.map((lib) => lib.driver.startup(hypervisor, ...lib.args)));
      return create_engine(hypervisor, canvas, ctx);
    }
  });
  return () => methods();
})();
export {
  src_default as default
};
